from onemod.config import Config, StageConfig
from onemod.main import load_pipeline, load_stage
from onemod.pipeline import Pipeline
from onemod.stage import Stage

__all__ = [
    "Config",
    "Pipeline",
    "Stage",
    "StageConfig",
    "load_pipeline",
    "load_stage",
]



"""Functions to run pipelines and stages locally."""

from typing import Any, Literal

from pydantic import validate_call

from onemod.backend.utils import (
    check_input_exists,
    check_method,
    collect_results,
)
from onemod.dtypes import UniqueList
from onemod.pipeline import Pipeline
from onemod.stage import Stage


@validate_call
def evaluate_local(
    model: Pipeline | Stage,
    method: Literal["run", "fit", "predict", "collect"],
    stages: UniqueList[str] | None = None,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    **kwargs,
) -> None:
    """Evaluate pipeline or stage method locally.

    Parameters
    ----------
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : {'run', 'fit', 'predict', 'collect}
        Name of method to evaluate.
    **kwargs
        Additional keyword arguments passed to stage methods. If `model`
        is a `Pipeline` instance, use format`stage={arg_name: arg_value}`.

    Pipeline Parameters
    -------------------
    stages : list of str, optional
        Names of stages to evaluate if `model` is a `Pipeline` instance.
        If None, evaluate all pipeline stages. Default is None.

    Stage Parameters
    ----------------
    subsets : dict, optional
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.

    """
    check_method(model, method)
    check_input_exists(model, stages)

    if isinstance(model, Pipeline):
        _evaluate_pipeline(model, method, stages, **kwargs)
    else:
        _evaluate_stage(model, method, subsets, paramsets, collect, **kwargs)


def _evaluate_pipeline(
    pipeline: Pipeline, method: str, stages: list[str] | None, **kwargs
) -> None:
    """Evaluate pipeline method locally."""
    for stage_name in pipeline.get_execution_order(stages):
        stage = pipeline.stages[stage_name]
        if method not in stage.skip:
            _evaluate_stage(stage, method, **kwargs.get(stage_name, {}))


def _evaluate_stage(
    stage: Stage,
    method: str,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    **kwargs,
) -> None:
    """Evaluate stage method locally."""
    if method == "collect":
        stage.collect()
    else:
        stage_method = stage.__getattribute__(f"_{method}")
        if stage.has_submodels:
            for subset, paramset in stage.get_submodels(subsets, paramsets):
                if subset is None:
                    stage_method(paramset=paramset, **kwargs)
                elif paramset is None:
                    stage_method(subset=subset, **kwargs)
                else:
                    stage_method(subset=subset, paramset=paramset, **kwargs)

            if collect_results(stage, method, subsets, paramsets, collect):
                stage.collect()
        else:
            stage_method(**kwargs)



"""Helper functions for unique lists and tuples."""

from typing import Annotated, Hashable, TypeVar

from pydantic import AfterValidator

T = TypeVar("T", bound=Hashable)


def unique_list(items: list[T]) -> list[T]:
    """Ensure all items in list are unique while preserving order."""
    return list(dict.fromkeys(items))


def update_unique_list(list1: list[T], list2: list[T]) -> list[T]:
    """Combine two lists, remove duplicates, and preserve order."""
    return list(dict.fromkeys(list1 + list2))


def unique_tuple(items: tuple[T, ...]) -> tuple[T, ...]:
    """Ensure all items in tuple are unique while preserving order."""
    return tuple(dict.fromkeys(items))


def update_unique_tuple(
    tuple1: tuple[T, ...], tuple2: tuple[T, ...]
) -> tuple[T, ...]:
    """Combine two tuples, remove duplicates, and preserve order."""
    return tuple(dict.fromkeys(tuple1 + tuple2))


UniqueList = Annotated[list[T], AfterValidator(unique_list)]
UniqueTuple = Annotated[tuple[T, ...], AfterValidator(unique_tuple)]



"""Functions to run pipelines and stages with Jobmon.

Examples
--------
Compute resources can be passed as a dictionary or a path to a resources
file (e.g., json, toml, yaml).

Required tool resources:

.. code-block:: yaml

    tool_resources:
      {cluster_name}:
        project: {proj_name}
        queue: {queue_name}

To test workflow on a dummy cluster (i.e., run workflow without running
tasks), use:

.. code-block:: yaml

    tool_resources:
      dummy:
        queue: null.q

Optional stage resources can be specified at the stage or stage + method
level:

    task_template_resources:
      {stage_name}:
        {cluster_name}:
            ...
      {stage_name}_{collect}:
        {cluster_name}:
            ...

See Jobmon documentation for additional resources and default values.

"""
# TODO: Optional stage-specific Python environments
# TODO: User-defined max_attempts
# TODO: Could dependencies be method specific?
# TODO: should we check resources format, minimum resources, cluster?

import sys
from pathlib import Path
from typing import Any, Literal

from jobmon.client.api import Tool
from jobmon.client.task import Task
from jobmon.client.task_template import TaskTemplate
from jobmon.client.workflow import Workflow
from pydantic import ConfigDict, validate_call

from onemod.backend.utils import (
    check_input_exists,
    check_method,
    collect_results,
)
from onemod.fsutils.config_loader import ConfigLoader
from onemod.pipeline import Pipeline
from onemod.stage import Stage


@validate_call
def evaluate_with_jobmon(
    model: Pipeline | Stage,
    method: Literal["run", "fit", "predict", "collect"],
    cluster: str,
    resources: Path | str | dict[str, Any],
    python: Path | str | None = None,
    stages: list[str] | None = None,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    task_and_template_prefix: str | None = None,
    max_attempts: int = 1,
    **kwargs,
) -> None:
    """Evaluate pipeline or stage method with Jobmon.

    Parameters
    ----------
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : {'run', 'fit', 'predict', 'collect'}
        Name of method to evalaute.
    cluster : str
        Cluster name.
    resources : dict, Path, or str
        Path to resources file or dictionary of compute resources.
    python : Path or str, optional
        Path to Python environment. If None, use sys.executable.
        Default is None.
    **kwargs
        Additional keyword arguments passed to stage methods. If `model`
        is a `Pipeline` instance, use format`stage={arg_name: arg_value}`.

    Pipeline Parameters
    -------------------
    stages : list of str, optional
        Names of stages to evaluate if `model` is a `Pipeline` instance.
        If None, evaluate pipeline stages. Default is None.

    Stage Parameters
    ----------------
    subsets : dict, optional
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.

    Jobmon Parameters
    -----------------
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name. Default is None,
        no prefix.
    max_attempts : int
        Maximum number of attempts for a task. Default is 1.

    """
    check_method(model, method)
    check_input_exists(model, stages)
    if python is None:
        python = str(sys.executable)

    resources_dict = get_resources(resources)
    workflow = create_workflow(model.name, method, cluster, resources_dict)
    add_tasks_to_workflow(
        model=model,
        workflow=workflow,
        method=method,
        resources=resources_dict,
        python=python,
        stages=stages,
        subsets=subsets,
        paramsets=paramsets,
        collect=collect,
        task_and_template_prefix=task_and_template_prefix,
        max_attempts=max_attempts,
        **kwargs,
    )
    run_workflow(workflow)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def add_tasks_to_workflow(
    model: Pipeline | Stage,
    workflow: Workflow,
    method: Literal["run", "fit", "predict", "collect"],
    resources: Path | str | dict[str, Any],
    python: Path | str | None = None,
    stages: list[str] | None = None,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    task_and_template_prefix: str | None = None,
    max_attempts: int = 1,
    external_upstream_tasks: list[Task] | None = None,
    **kwargs,
) -> None:
    """Add Pipeline tasks to an existing Jobmon Workflow.

    Note that this is a publically available function, be careful of
    breaking changes to the API or functionality.

    Parameters
    ----------
    workflow : Workflow
        Instantiated Jobmon workflow. Add new tasks to an existing Jobmon
        workflow rather than creating a new workflow. Does not run the
        workflow, only adds the tasks.
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : {'run', 'fit', 'predict', 'collect'}
        Name of method to evalaute.
    resources : dict, Path, or str
        Path to resources file or dictionary of compute resources.
    python : Path or str, optional
        Path to Python environment. If None, use sys.executable.
        Default is None.
    **kwargs
        Additional keyword arguments passed to stage methods. If `model`
        is a `Pipeline` instance, use format`stage={arg_name: arg_value}`.

    Pipeline Parameters
    -------------------
    stages : list of str, optional
        Names of stages to evaluate if `model` is a `Pipeline` instance.
        If None, evaluate pipeline stages. Default is None.

    Stage Parameters
    ----------------
    subsets : dict, optional
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.

    Jobmon Parameters
    -----------------
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name. Default is None,
        no prefix.
    max_attempts : int
        Maximum number of attempts for a task. Default is 1.
    external_upstream_tasks : list, optional
        List of Jobmon tasks external to the OneMod Stages or Pipeline that
        should be treated as upstream dependencies of the new tasks. Default
        is no external upstream tasks.

    """
    check_method(model, method)
    check_input_exists(model, stages)
    if python is None:
        python = str(sys.executable)

    resources_dict = get_resources(resources)
    tasks = get_tasks(
        model=model,
        method=method,
        tool=workflow.tool,
        resources=resources_dict,
        python=python,
        stages=stages,
        subsets=subsets,
        paramsets=paramsets,
        collect=collect,
        task_and_template_prefix=task_and_template_prefix,
        max_attempts=max_attempts,
        external_upstream_tasks=external_upstream_tasks,
        **kwargs,
    )
    workflow.add_tasks(tasks)


def get_resources(resources: Path | str | dict[str, Any]) -> dict[str, Any]:
    """Get dictionary of compute resources.

    Parameters
    ----------
    resources : Path, str, or dict
        Path to resources file or dictionary of compute resources.

    Returns
    -------
    dict
        Dictionary of compute resources.

    """
    if isinstance(resources, (Path, str)):
        config_loader = ConfigLoader()
        return config_loader.load(Path(resources))
    return resources


def create_workflow(
    name: str,
    method: Literal["run", "fit", "predict", "collect"],
    cluster: str,
    resources: dict[str, Any],
) -> Workflow:
    """Create and return workflow.

    Parameters
    ----------
    name : str
        Pipeline or stage name.
    method : str
        Name of method being evaluated.
    cluster : str
        Cluster name.
    resources : dict
        Dictionary of compute resources.

    Returns
    -------
    Workflow
        Jobmon workflow.

    """
    tool = get_tool(name, method, cluster, resources)
    workflow = tool.create_workflow(name=f"{name}_{method}")
    return workflow


def get_tool(
    name: str, method: str, cluster: str, resources: dict[str, Any]
) -> Tool:
    """Get Jobmon tool.

    Parameters
    ----------
    name : str
        Pipeline or stage name.
    method : str
        Name of method to evaluate.
    cluster : str
        Cluster name.
    resources : dict
        Dictionary of compute resources.

    Returns
    -------
    Tool
        Jobmon tool.

    """
    tool = Tool(name=f"{name}_{method}")
    tool.set_default_cluster_name(cluster)
    tool.set_default_compute_resources_from_dict(
        cluster, resources["tool_resources"][cluster]
    )
    return tool


def get_tasks(
    model: Pipeline | Stage,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    stages: list[str] | None,
    subsets: dict[str, Any | list[Any]] | None,
    paramsets: dict[str, Any | list[Any]] | None,
    collect: bool | None,
    task_and_template_prefix: str | None,
    max_attempts: int,
    external_upstream_tasks: list[Task] | None = None,
    **kwargs,
) -> list[Task]:
    """Get Jobmon tasks.

    Parameters
    ----------
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    **kwargs
        Additional keyword arguments passed to stage methods.

    Pipeline Parameters
    -------------------
    stages : list of str or None
        Name of stages to evaluate if `model` is a pipeline instance. If
        None, evaluate all pipeline stages.

    Stage Parameters
    ----------------
    subsets : dict or None
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets.
    paramsets : dict or None
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets.
    collect : bool or None
        Whether to collect submodel results if `model` is a `Stage`
        instance.

    Jobmon Parameters
    -----------------
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name.
    max_attempts : int
        Maximum number of attempts for a task.
    external_upstream_tasks : list, optional
        List of Jobmon tasks external to the OneMod Stages or Pipeline that
        should be treated as upstream dependencies of the new tasks. Default
        None, no external upstreams.

    Returns
    -------
    list of Task
        List of Jobmon tasks.

    """
    if isinstance(model, Pipeline):
        return get_pipeline_tasks(
            pipeline=model,
            method=method,
            tool=tool,
            resources=resources,
            python=python,
            stages=stages,
            external_upstream_tasks=external_upstream_tasks,
            task_and_template_prefix=task_and_template_prefix,
            max_attempts=max_attempts,
            **kwargs,
        )
    return get_stage_tasks(
        stage=model,
        method=method,
        tool=tool,
        resources=resources,
        python=python,
        task_and_template_prefix=task_and_template_prefix,
        max_attempts=max_attempts,
        subsets=subsets,
        paramsets=paramsets,
        collect=collect,
        upstream_tasks=external_upstream_tasks,
        **kwargs,
    )


def get_pipeline_tasks(
    pipeline: Pipeline,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    stages: list[str] | None,
    external_upstream_tasks: list[Task] | None,
    task_and_template_prefix: str | None,
    max_attempts: int,
    **kwargs,
) -> list[Task]:
    """Get pipeline stage tasks.

    Parameters
    ----------
    pipeline : Pipeline
        Pipeline instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    stages : list of str or None
        Name of stages to evaluate. If None, evaluate all pipeline
        stages.
    external_upstream_tasks : list, optional
        List of Jobmon tasks external to the OneMod Stages or Pipeline that
        should be treated as upstream dependencies of the new tasks.
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name.
    max_attempts : int
        Maximum number of attempts for a task.
    **kwargs
        Additional keyword arguments passed to stage methods.

    Returns
    -------
    list of Task
        List of pipeline stage tasks.

    """
    tasks = []
    task_dict: dict[str, list[Task]] = {}
    task_dict["external"] = external_upstream_tasks or []

    for stage_name in pipeline.get_execution_order(stages):
        stage = pipeline.stages[stage_name]
        if method not in stage.skip:
            upstream_tasks = get_upstream_tasks(
                stage, method, pipeline.stages, task_dict, stages
            )
            task_dict[stage_name] = get_stage_tasks(
                stage=stage,
                method=method,
                tool=tool,
                resources=resources,
                python=python,
                task_and_template_prefix=task_and_template_prefix,
                max_attempts=max_attempts,
                upstream_tasks=upstream_tasks,
                **kwargs,
            )
            tasks.extend(task_dict[stage_name])

    return tasks


def get_upstream_tasks(
    stage: Stage,
    method: str,
    stage_dict: dict[str, Stage],
    task_dict: dict[str, list[Task]],
    stages: list[str] | None,
) -> list[Task]:
    """Get upstream tasks for current stage.

    Parameters
    ----------
    stage : Stage
        Current stage instance.
    method : str
        Name of method to evaluate.
    stage_dict : dict
        Dictionary of all pipeline stages.
    task_dict : dict
        Dictionary of all upstream stage tasks.
    stages : list of str or None
        Names of all pipeline stages being evaluated. If None, assume
        all stages are being evaluated.

    Returns
    -------
    list of Task
        Upstream stage tasks for current stage.

    Notes
    -----
    * Only include tasks corresponding to the current stage's
      dependencies that are included in `stages`.
    * If an upstream stage has submodels and `method` is in the
      upstream's `collect_after`, only include the task corresponding to
      the upstream's `collect` method.
    * If there are no upstream tasks, add any external tasks as upstream.

    """
    upstream_tasks = []

    for upstream_name in stage.dependencies:
        if stages is not None and upstream_name not in stages:
            # upstream stage is not being evaluated
            continue

        upstream_stage = stage_dict[upstream_name]
        if method not in upstream_stage.skip:
            if (
                upstream_stage.has_submodels
                and method in upstream_stage.collect_after
            ):
                # only include task corresponding to 'collect' method
                upstream_tasks.append(task_dict[upstream_name][-1])
            else:
                upstream_tasks.extend(task_dict[upstream_name])

    # if there are no upstream tasks, add external upstream tasks
    if not upstream_tasks:
        upstream_tasks = task_dict.get("external", [])

    return upstream_tasks


def get_stage_tasks(
    stage: Stage,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    task_and_template_prefix: str | None,
    max_attempts: int,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    upstream_tasks: list[Task] | None = None,
    **kwargs,
) -> list[Task]:
    """Get stage tasks.

    Parameters
    ----------
    stage : Stage
        Stage instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    subsets : dict, optional
        Submodel data subsets to evaluate. If None, evaluate all data
        subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate. If None, evaluate all
        parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.
    upstream_tasks : list of Task or None, optional
        List of upstream stage tasks. Default is None.
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name.
    max_attempts : int
        Maximum number of attempts for a task.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    list of Task
        List of stage tasks.

    """
    if upstream_tasks is None:
        upstream_tasks = []

    entrypoint = str(Path(python).parent / "onemod")
    config_path = str(stage.dataif.get_path("config"))
    submodel_args = get_submodel_args(stage, method, subsets, paramsets)

    task_template = get_task_template(
        stage.name,
        method,
        tool,
        resources,
        list(submodel_args.keys()),
        task_and_template_prefix=task_and_template_prefix,
        **kwargs,
    )

    task_name = (
        f"{task_and_template_prefix}_{stage.name}_{method}"
        if task_and_template_prefix
        else f"{stage.name}_{method}"
    )
    if submodel_args:
        tasks = task_template.create_tasks(
            name=task_name,
            upstream_tasks=upstream_tasks,
            max_attempts=max_attempts,
            entrypoint=entrypoint,
            config=config_path,
            method=method,
            stages=stage.name,
            **{**submodel_args, **kwargs},
        )
    else:
        tasks = [
            task_template.create_task(
                name=task_name,
                upstream_tasks=upstream_tasks,
                max_attempts=max_attempts,
                entrypoint=entrypoint,
                config=config_path,
                method=method,
                stages=stage.name,
                **kwargs,
            )
        ]

    if collect_results(stage, method, subsets, paramsets, collect):
        tasks.extend(
            get_stage_tasks(
                stage=stage,
                method="collect",
                tool=tool,
                resources=resources,
                python=python,
                task_and_template_prefix=task_and_template_prefix,
                max_attempts=max_attempts,
                upstream_tasks=tasks,
            )
        )

    return tasks


def get_submodel_args(
    stage: Stage,
    method: str,
    subsets: dict[str, Any | list[Any]] | None,
    paramsets: dict[str, Any | list[Any]] | None,
) -> dict[str, list[str]]:
    """Get dictionary of subset and/or paramset values.

    If stage has submodels and `method` is not 'collect', additional
    args for 'subsets' and/or 'paramsets' are included in the command
    template.

    Parameters
    ----------
    stage : Stage
        Stage instance.
    method : str
        Method being evaluated.
    subsets : dict or None.
        Submodel data subsets to evaluate. If None, evaluate all data
        subsets.
    paramsets : dict or None
        Submodel parameter sets to evaluate. If None, evaluate all
        parameter sets.

    Returns
    -------
    dict
        Dictionary of subset and/or paramset values as strings.

    """
    submodel_args = {}
    if stage.has_submodels and method != "collect":
        # Get data subsets
        if (filtered_subsets := stage.subsets) is not None:
            if subsets is not None:
                filtered_subsets = stage.get_subset(filtered_subsets, subsets)
            submodel_args["subsets"] = [
                str(subset)
                for subset in filtered_subsets.to_dict(orient="records")
            ]

        # Get parameter sets
        if (filtered_paramsets := stage.paramsets) is not None:
            if paramsets is not None:
                filtered_paramsets = stage.get_subset(
                    filtered_paramsets, paramsets
                )
            submodel_args["paramsets"] = [
                str(paramset)
                for paramset in filtered_paramsets.to_dict(orient="records")
            ]

    return submodel_args


def get_task_template(
    stage_name: str,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    submodel_args: list[str],
    task_and_template_prefix: str | None,
    **kwargs,
) -> TaskTemplate:
    """Get stage task template.

    If the Jobmon Tool already has an active task template with the same
    name, use that task template.

    Parameters
    ----------
    stage_name : str
        Stage name.
    method : str
        Name of method being evaluated.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    submodel_args : list of str
        List including 'subsets' and/or 'paramsets'.
    task_and_template_prefix : str, optional
        Optional prefix to append to task/template name.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    TaskTemplate
        Stage task template.

    """
    template_name = (
        f"{task_and_template_prefix}_{stage_name}_{method}"
        if task_and_template_prefix
        else f"{stage_name}_{method}"
    )

    task_template = tool.get_task_template(
        template_name=template_name,
        command_template=get_command_template(method, submodel_args, **kwargs),
        op_args=["entrypoint"],
        task_args=["config", "method", "stages"] + list(kwargs.keys()),
        node_args=submodel_args,
    )

    task_resources = get_task_resources(
        resources, tool.default_cluster_name, stage_name, method
    )
    if task_resources:
        task_template.set_default_compute_resources_from_dict(
            tool.default_cluster_name, task_resources
        )

    return task_template


def get_command_template(
    method: str, submodel_args: list[str], **kwargs
) -> str:
    """Get stage command template.

    All stages methods are called via `onemod.main.evaluate()`. If stage
    has submodels and `method` is not 'collect', additional args for
    'subsets' and/or 'paramsets' are included in the command template.

    Parameters
    ----------
    method : str
        Name of method being evaluated.
    submodel_args : list of str
        List including 'subsets' and/or 'paramsets'.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    str
        Stage command template.

    # TODO: collapse into one list

    """
    command_template = (
        "{entrypoint} --config {config} --method {method} --stages {stages}"
    )

    if method != "collect":
        for key, value in kwargs.items():
            if isinstance(value, (dict, list, set, tuple)):
                command_template += f" --{key} '{{{key}}}'"
            else:
                command_template += f" --{key} {{{key}}}"

        for arg in submodel_args:
            command_template += f" --{arg} '{{{arg}}}'"

    return command_template


def get_task_resources(
    resources: dict[str, Any], cluster: str, stage_name: str, method: str
) -> dict[str, Any]:
    """Get task-specific resources.

    Parameters
    ----------
    resources : dict
        Dictionary of compute resources.
    cluster : str
        Cluster name.
    stage_name : str
        Stage name.
    method : str
        Name of method being evaluated.

    Returns
    -------
    dict
        Task-specific resources.

    """
    task_resources = resources.get("task_template_resources", {})
    stage_resources = task_resources.get(stage_name, {})
    method_resources = task_resources.get(f"{stage_name}_{method}", {})
    return {
        **stage_resources.get(cluster, {}),
        **method_resources.get(cluster, {}),
    }


def run_workflow(workflow: Workflow) -> None:
    """Run workflow.

    Parameters
    ----------
    workflow : Workflow
        Jobmon workflow to run.

    """
    workflow.bind()
    print(f"Starting workflow {workflow.workflow_id}")
    status = workflow.run()
    if status != "D":
        raise ValueError(f"Workflow {workflow.workflow_id} failed")
    else:
        print(f"Workflow {workflow.workflow_id} finished")



"""Pipeline class."""

from __future__ import annotations

import json
import logging
from collections import deque
from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel

from onemod.config import Config
from onemod.serialization import serialize
from onemod.stage import Stage
from onemod.utils.decorators import computed_property
from onemod.validation import ValidationErrorCollector, handle_error

logger = logging.getLogger(__name__)


class Pipeline(BaseModel):
    """Pipeline class.

    Parameters
    ----------
    name : str
        Pipeline name.
    directory : Path or str
        Path to pipeline directory.
    config : Config or dict, optional
        Pipeline configuration.
    groupby_data : Path or str, optional
        Path to data file used to create stage data subsets. Must
        contain all columns included in stage ``groupby`` attributes.

    """

    name: str
    directory: Path
    config: Config = Config()
    groupby_data: Path | None = None
    _stages: dict[str, Stage] = {}

    @computed_property
    def stages(self) -> dict[str, Stage]:
        """Pipeline stages."""
        return self._stages

    @computed_property
    def dependencies(self) -> dict[str, list[str]]:
        """Stage dependencies."""
        return {
            stage.name: stage.dependencies for stage in self.stages.values()
        }

    @classmethod
    def from_json(cls, config_path: Path | str) -> Pipeline:
        """Load pipeline from JSON file.

        Parameters
        ----------
        config_path : Path or str
            Path to pipeline config file.

        Returns
        -------
        Pipeline
            Pipeline instance.

        """
        with open(config_path, "r") as file:
            config = json.load(file)

        stages = config.pop("stages", {})

        pipeline = cls(**config)

        if stages:
            from onemod.main import load_stage

            pipeline.add_stages(
                [load_stage(config_path, stage) for stage in stages]
            )

        return pipeline

    def to_json(self, config_path: Path | str | None = None) -> None:
        """Save pipeline as JSON file.

        Parameters
        ----------
        config_path : Path or str, optional
            Where to save config file. If None, file is saved at
            pipeline.directory / (pipeline.name + ".json"). Default is
            None.

        """
        config_path = config_path or self.directory / (self.name + ".json")
        with open(config_path, "w") as file:
            file.write(
                self.model_dump_json(
                    indent=2, exclude_none=True, serialize_as_any=True
                )
            )

    def add_stages(self, stages: list[Stage]) -> None:
        """Add stages to pipeline.

        Parameters
        ----------
        stages : list of Stage
            Stages to add to the pipeline.

        """
        for stage in stages:
            self.add_stage(stage)

    def add_stage(self, stage: Stage) -> None:
        """Add stage to pipeline.

        Parameters
        ----------
        stage : Stage
            Stage to add to the pipeline.

        """
        if stage.name in self.stages:
            raise ValueError(f"Stage '{stage.name}' already exists")

        stage.config.add_pipeline_config(self.config)
        stage.set_dataif(self.directory / (self.name + ".json"))
        stage.set_output()

        self._stages[stage.name] = stage

    def get_execution_order(self, stages: list[str] | None = None) -> list[str]:
        """Get stages sorted in execution order.

        Use Kahn's algorithm to find the topoligical order of the
        stages, ensuring no cycles.

        Parameters
        ----------
        stages: list of str, optional
            Name of stages to sort. If None, sort all pipeline stages.
            Default is None.

        Returns
        -------
        list of str
            Stages sorted in execution order.

        Raises
        ------
        ValueError
            If cycle detected in DAG.

        """
        # TODO: What if stages have a gap? For example, pipeline has
        # Rover -> SPxMod -> KReg, but `stages` only includes Rover and
        # Kreg. KReg will be run on outdated SPxMod results (if they
        # exist).
        reverse_graph: dict[str, list[str]] = {
            stage: [] for stage in self.dependencies
        }
        in_degree = {stage: 0 for stage in self.dependencies}
        for stage, deps in self.dependencies.items():
            for dep in deps:
                reverse_graph[dep].append(stage)
                in_degree[stage] += 1

        queue = deque([stage for stage, deg in in_degree.items() if deg == 0])
        topological_order = []
        visited = set()

        while queue:
            stage = queue.popleft()
            topological_order.append(stage)
            visited.add(stage)

            # Reduce the in-degree of downstream stages
            for downstream_dep in reverse_graph[stage]:
                in_degree[downstream_dep] -= 1
                if in_degree[downstream_dep] == 0:
                    queue.append(downstream_dep)

        # If there is a cycle, the topological order will not include all stages
        if len(topological_order) != len(self.dependencies):
            unvisited = set(self.dependencies) - visited
            raise ValueError(
                f"Cycle detected! Unable to process the following stages: {unvisited}"
            )

        if stages:
            return [stage for stage in topological_order if stage in stages]
        return topological_order

    def validate_dag(self, collector: ValidationErrorCollector) -> None:
        """Validate that the DAG structure is correct."""
        for stage_name, dependencies in self.dependencies.items():
            for dep in dependencies:
                if dep not in self._stages:
                    handle_error(
                        stage_name,
                        "DAG validation",
                        ValueError,
                        f"Upstream dependency '{dep}' is not defined.",
                        collector,
                    )

                if dep == stage_name:
                    handle_error(
                        stage_name,
                        "DAG validation",
                        ValueError,
                        "Stage cannot depend on itself.",
                        collector,
                    )

            if len(dependencies) != len(set(dependencies)):
                handle_error(
                    stage_name,
                    "DAG validation",
                    ValueError,
                    "Duplicate dependencies found.",
                    collector,
                )

        try:
            self.get_execution_order()
        except ValueError as e:
            handle_error(
                "Pipeline", "DAG validation", ValueError, str(e), collector
            )

    def build(self) -> None:
        """Assemble pipeline, perform build-time validation, and save to JSON."""
        collector = ValidationErrorCollector()

        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        for stage in self.stages.values():
            stage.build(collector, self.groupby_data)

        self.validate_dag(collector)

        if collector.has_errors():
            self.save_validation_report(collector)
            collector.raise_errors()

        self.to_json(self.directory / (self.name + ".json"))

    def save_validation_report(
        self, collector: ValidationErrorCollector
    ) -> None:
        validation_dir = self.directory / "validation"
        validation_dir.mkdir(exist_ok=True)
        report_path = validation_dir / "validation_report.json"
        serialize(collector.errors, report_path)  # type: ignore[arg-type]

    def evaluate(
        self,
        method: Literal["run", "fit", "predict", "collect"],
        stages: list[str] | None,
        backend: Literal["local", "jobmon"],
        cluster: str | None,
        resources: Path | str | dict[str, Any] | None,
        python: Path | str | None,
        **kwargs,
    ) -> None:
        """Evaluate pipeline method."""
        if backend == "jobmon":
            from onemod.backend.jobmon_backend import evaluate_with_jobmon

            if cluster is None:
                raise ValueError("Jobmon backend requires cluster name")
            if resources is None:
                raise ValueError("Jobmon backend requires compute resources")

            evaluate_with_jobmon(
                model=self,
                method=method,
                stages=stages,
                cluster=cluster,
                resources=resources,
                python=python,
                **kwargs,
            )
        else:
            from onemod.backend.local_backend import evaluate_local

            evaluate_local(model=self, method=method, stages=stages, **kwargs)

    def run(
        self,
        stages: list[str] | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Run pipeline stages.

        All stage submodels are run and submodel results are collected.

        Parameters
        ----------
        stages : list of str, optional
            Names of stages to run. If None, run all pipeline stages.
            Default is None.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage methods. Use
            format ``stage={arg_name: arg_value}``.

        Jobmon Parameters
        ----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        python : Path, or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "run", stages, backend, cluster, resources, python, **kwargs
        )

    def fit(
        self,
        stages: list[str] | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Fit pipeline stages.

        All stage submodels are fit and submodel results are collected.

        Parameters
        ----------
        stages : list of str, optional
            Names of stages to fit. If None, fit entire pipeline.
            Default is None.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage methods. Use
            format ``stage={arg_name: arg_value}``.

        Jobmon Parameters
        -----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        python : Path, or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "fit", stages, backend, cluster, resources, python, **kwargs
        )

    def predict(
        self,
        stages: list[str] | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Create predictions for pipeline stages.

        Predictions are made for all stage submodels and submodel
        results are collected.

        Parameters
        ----------
        stages : list of str, optional
            Names of stages to create predictions for. If None, create
            predictions for the entire pipeline. Default is None.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage methods. Use
            format ``stage={arg_name: arg_value}``.

        Jobmon Parameters
        -----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        python : Path, or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "predict", stages, backend, cluster, resources, python, **kwargs
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(name={self.name},"
            f" stages={list(self.stages.values())})"
        )



from pathlib import Path
from typing import Any, Literal, Mapping

import pandas as pd
import polars as pl

from onemod.fsutils.io import DataIO, dataio_dict


class DataLoader:
    """Handles loading and dumping of data files with optional lazy loading,
    column selection, and subset filtering."""

    io_dict: dict[str, DataIO] = dataio_dict

    def load(
        self,
        path: Path,
        return_type: Literal[
            "pandas_dataframe", "polars_dataframe", "polars_lazyframe"
        ] = "pandas_dataframe",
        columns: list[str] | None = None,
        subset: Mapping[str, Any | list[Any]] | None = None,
        **options,
    ) -> pd.DataFrame | pl.DataFrame | pl.LazyFrame:
        """Load data with lazy loading and subset filtering. Polars and
        Pandas options available for the type of the returned data object."""

        if path.suffix not in self.io_dict:
            raise ValueError(f"Unsupported data format for '{path.suffix}'")

        if return_type == "pandas_dataframe":
            pandas_df = self.io_dict[path.suffix].load_eager(path, **options)
            assert isinstance(
                pandas_df, pd.DataFrame
            ), "Expected a pandas DataFrame"

            if columns:
                pandas_df = pandas_df[columns]

            if subset:
                for col, values in subset.items():
                    pandas_df = pandas_df[
                        pandas_df[col].isin(
                            values if isinstance(values, list) else [values]
                        )
                    ]
                    pandas_df.reset_index(drop=True, inplace=True)

            return pandas_df
        elif return_type in ["polars_dataframe", "polars_lazyframe"]:
            polars_lf = self.io_dict[path.suffix].load_lazy(path, **options)

            if columns:
                polars_lf = polars_lf.select(columns)

            if subset:
                for col, values in subset.items():
                    polars_lf = polars_lf.filter(
                        pl.col(col).is_in(
                            values if isinstance(values, list) else [values]
                        )
                    )

            if return_type == "polars_dataframe":
                return polars_lf.collect()
            elif return_type == "polars_lazyframe":
                return polars_lf
        else:
            raise ValueError(
                "Return type must be one of 'polars_dataframe', 'polars_lazyframe', or 'pandas_dataframe'"
            )

    def dump(
        self,
        obj: pd.DataFrame | pl.DataFrame | pl.LazyFrame,
        path: Path,
        **options,
    ) -> None:
        """Save data file."""
        if path.suffix not in self.io_dict:
            raise ValueError(f"Unsupported data format for '{path.suffix}'")

        self.io_dict[path.suffix].dump(obj, path, **options)

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"



from onemod.validation.error_handling import (
    ValidationErrorCollector,
    ValidationErrorReport,
    handle_error,
    validation_context,
)

__all__ = [
    "ValidationErrorCollector",
    "ValidationErrorReport",
    "handle_error",
    "validation_context",
]



from pathlib import Path
from typing import Any, Literal, Mapping

import pandas as pd
import polars as pl

from onemod.fsutils.config_loader import ConfigLoader
from onemod.fsutils.data_loader import DataLoader
from onemod.fsutils.path_manager import PathManager


class DataInterface(PathManager):
    """Unified interface for loading and dumping data and config files."""

    def __init__(self, **paths: Path | str) -> None:
        super().__init__(**paths)
        self.data_loader = DataLoader()
        self.config_loader = ConfigLoader()

    def load(
        self,
        *fparts: str,
        key: str | None = None,
        return_type: Literal[
            "pandas_dataframe", "polars_dataframe", "polars_lazyframe"
        ] = "pandas_dataframe",
        columns: list[str] | None = None,
        subset: Mapping[str, Any | list[Any]] | None = None,
        **options,
    ) -> Any:
        """Load data or config files based on file type and key.

        Parameters
        ----------
        return_type : {'pandas_dataframe', 'polars_dataframe', 'polars_lazyframe'}, optional
            Return type of loaded data object, applicable only for data files.
        columns : list of str, optional
            Specific columns to load, applicable only for data files.
        subset : dict, optional
            Subset of column values for filtering, applicable only for
            data files.
        options : dict
            Additional options for loading.

        Returns
        -------
        Any
            Loaded data or config file.
        """
        path = self.get_full_path(*fparts, key=key)
        if path.suffix in self.config_loader.io_dict.keys():
            return self.config_loader.load(path, **options)
        elif path.suffix in self.data_loader.io_dict.keys():
            return self.data_loader.load(
                path,
                return_type=return_type,
                columns=columns,
                subset=subset,
                **options,
            )
        else:
            raise ValueError(f"Unsupported file format for '{path.suffix}'")

    def dump(
        self, obj: Any, *fparts: str, key: str | None = None, **options
    ) -> None:
        """Dump data or config files based on object type and key."""
        path = self.get_full_path(*fparts, key=key)
        if isinstance(obj, (pd.DataFrame, pl.DataFrame, pl.LazyFrame)):
            return self.data_loader.dump(obj, path, **options)
        else:
            return self.config_loader.dump(obj, path, **options)



from pathlib import Path
from typing import Any

from onemod.fsutils.io import ConfigIO, configio_dict


class ConfigLoader:
    """Handles loading and dumping of configuration files and serialized models."""

    io_dict: dict[str, ConfigIO] = configio_dict

    def load(self, path: Path, **options) -> Any:
        """Load a config file or serialized model."""
        if path.suffix not in self.io_dict:
            raise ValueError(f"Unsupported config format for '{path.suffix}'")

        return self.io_dict[path.suffix].load(path, **options)

    def dump(self, obj: Any, path: Path, **options) -> None:
        """Save a config or model object."""
        if path.suffix not in self.io_dict:
            raise ValueError(f"Unsupported config format for '{path.suffix}'")

        self.io_dict[path.suffix].dump(obj, path, **options)

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"



from onemod.dtypes.column_spec import ColumnSpec
from onemod.dtypes.data import Data
from onemod.dtypes.filepath import FilePath
from onemod.dtypes.unique_sequence import UniqueList, UniqueTuple

__all__ = ["ColumnSpec", "Data", "FilePath", "UniqueList", "UniqueTuple"]



"""Input/output classes.

Notes
-----
* Classes to organize stage input and output
* Input and output treated like a dictionary
* Provides validation
* No need to create stage-specific subclasses

"""
# TODO: Use collector with check_cycles and check_types?
# TODO: Error messages weird for check_cycles/check_types?

from abc import ABC
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict, model_serializer

from onemod.dtypes import Data
from onemod.dtypes.unique_sequence import unique_list
from onemod.validation.error_handling import (
    ValidationErrorCollector,
    handle_error,
)


class IO(BaseModel, ABC):
    """Stage input/output base class."""

    model_config = ConfigDict(frozen=True)

    stage: str
    items: dict[str, Data] = {}

    @model_serializer
    def serialize_io(self) -> dict[str, dict[str, Any]]:
        # Simplify output to config files
        input_dict: dict[str, dict[str, Any]] = {}
        for item_name, item_value in self.items.items():
            input_dict[item_name] = item_value.model_dump(
                exclude_none=True, serialize_as_any=True
            )
        return input_dict

    def get(self, item_name: str, default: Any = None) -> Any:
        return self.items.get(item_name, default)

    def __getitem__(self, item_name: str) -> Data:
        return self.items[item_name]

    def __contains__(self, item_name: str) -> bool:
        return item_name in self.items


class Input(IO):
    """Stage input class."""

    required: dict[str, Data] = {}
    optional: dict[str, Data] = {}

    @property
    def dependencies(self) -> list[str]:
        return unique_list(
            [item.stage for item in self.items.values() if item.stage]
        )

    def model_post_init(self, *args, **kwargs) -> None:
        if len(set(self.required).intersection(self.optional)):
            raise ValueError("Required and optional input cannot share keys")
        if self.items:
            # Note: silently removes anything not in required or optional
            for item_name in list(self.items):
                if item_name not in list(self.required) + list(self.optional):
                    del self.items[item_name]
            self._check_cycles()
            self._check_types()

    def update(self, items: dict[str, Data | Path | str]) -> None:
        # Note: silently ignores anything not in required or optional
        data_items = {}
        for item_name, item_value in items.items():
            if item_name in list(self.required) + list(self.optional):
                if isinstance(item_value, Data):
                    data_items[item_name] = item_value
                elif isinstance(item_value, (Path, str)):
                    data_items[item_name] = Input.path_to_data(item_value)
                else:
                    raise TypeError(
                        f"Invalid input item: {item_name}={item_value}"
                    )

        self._check_cycles(data_items)
        self._check_types(data_items)

        for item_name, item_value in data_items.items():
            self.items[item_name] = item_value

    def remove(self, item: str) -> None:
        if item in self.items:
            del self.items[item]

    def clear(self) -> None:
        self.items.clear()

    def check_missing(
        self,
        items: dict[str, Data] | None = None,
        collector: ValidationErrorCollector | None = None,
    ) -> None:
        """Check stage input items have been defined.

        Parameters
        ----------
        items : dict of str: Data, optional
            Input items to check. If None, check all stage input items.
            Default is None.

        Raises
        ------
        KeyError
            If any stage input items have not been defined.

        """
        items = items or self.items
        missing_items = [
            item_name for item_name in self.required if item_name not in items
        ]
        if missing_items:
            handle_error(
                self.stage,
                "Data validation",
                KeyError,
                f"Stage '{self.stage}' missing required input: {missing_items}",
                collector,
            )

    def check_exists(
        self,
        item_names: list[str] | None = None,
        upstream_stages: list[str] | None = None,
        collector: ValidationErrorCollector | None = None,
    ) -> None:
        """Check stage input items exist.

        Parameters
        ----------
        item_names : list of str, optional
            Names of input items to check. If None, check all input
            path items and all input data items in `upstream_stages`.
            Default is None.
        upstream_stages : list of str, optional
            Names of upstream stages to check input items from. If None,
            check input items from all upstream stages.

        Raises
        ------
        FileNotFoundError
            If any stage input items do not exist.

        """
        if item_names is None:
            item_names = list(self.items)
        if upstream_stages is None:
            upstream_stages = self.dependencies

        missing_items: dict[str, str | None] = {}
        for item_name in item_names:
            item_value = self.__getitem__(item_name)
            if item_value.stage in upstream_stages:
                if (item_path := item_value.path) is None:
                    missing_items[item_name] = None
                elif not item_path.exists():
                    missing_items[item_name] = str(item_path)
            else:
                continue

        if missing_items:
            handle_error(
                self.stage,
                "Data validation",
                FileNotFoundError,
                f"Stage '{self.stage}' input items do not exist: {missing_items}",
                collector,
            )

    def _check_cycles(self, items: dict[str, Data] | None = None) -> None:
        cycles = []
        items = items or self.items
        for item_name, item_value in items.items():
            try:
                self._check_cycle(item_name, item_value)
            except ValueError:
                cycles.append(item_name)
        if cycles:
            raise ValueError(
                f"Circular dependencies for {self.stage} input: {cycles}"
            )

    def _check_cycle(self, item_name: str, item_value: Data) -> None:
        if item_value.stage == self.stage:
            raise ValueError(
                f"Circular dependency for {self.stage} input: {item_name}"
            )

    def _check_types(self, items: dict[str, Data] | None = None) -> None:
        invalid_items = []
        items = items or self.items
        for item_name, item_value in items.items():
            try:
                self._check_type(item_name, item_value)
            except TypeError:
                invalid_items.append(item_name)
        if invalid_items:
            raise TypeError(
                f"Invalid types for {self.stage} input: {invalid_items}"
            )

    def _check_type(self, item_name: str, item_value: Data) -> None:
        if item_name in list(self.required):
            expected = self.required[item_name]
        elif item_name in list(self.optional):
            expected = self.optional[item_name]
        else:
            return

        if item_value.format != expected.format:
            raise TypeError(f"Invalid type for {self.stage} input: {item_name}")

    @staticmethod
    def path_to_data(path: Path | str) -> Data:
        if isinstance(path, str):
            path = Path(path)
        return Data(format=path.suffix[1:] or "directory", path=path)

    def __getitem__(self, item_name: str) -> Data:
        if item_name not in self.items:
            if item_name in list(self.required) + list(self.optional):
                raise ValueError(
                    f"{self.stage} input '{item_name}' has not been set"
                )
            raise KeyError(f"{self.stage} does not contain input '{item_name}'")
        return self.items[item_name]

    def __setitem__(
        self, item_name: str, item_value: Data | Path | str
    ) -> None:
        if item_name in list(self.required) + list(self.optional):
            if isinstance(item_value, (Path, str)):
                item_value = self.path_to_data(item_value)
            self._check_cycle(item_name, item_value)
            self._check_type(item_name, item_value)
            self.items[item_name] = item_value


class Output(IO):
    """Stage output class."""

    directory: Path
    items: dict[str, Data] = {}

    def model_post_init(self, *args, **kwargs) -> None:
        for item_name, item_value in self.items.items():
            item_value.stage = self.stage
            item_value.path = self.directory / (
                item_name
                if (item_format := item_value.format) == "directory"
                else f"{item_name}.{item_format}"
            )

    def check_exists(
        self,
        item_names: list[str] | None = None,
        collector: ValidationErrorCollector | None = None,
    ) -> None:
        # TODO: add method arg
        if item_names is None:
            item_names = list(self.items)

        missing_items: dict[str, str | None] = {}
        for item_name in item_names:
            item_value = self.__getitem__(item_name)
            if (item_path := item_value.path) is None:
                missing_items[item_name] = None
            elif not item_path.exists():
                missing_items[item_name] = str(item_path)

        if missing_items:
            handle_error(
                self.stage,
                "Data validation",
                FileNotFoundError,
                f"Stage '{self.stage}' output items do not exist: {missing_items}",
                collector,
            )

    def __getitem__(self, item_name: str) -> Data:
        if item_name not in self.items:
            raise KeyError(
                f"Stage '{self.stage}' does not contain output '{item_name}'"
            )
        return self.items[item_name]



"""Utility functions used by all backends."""
# TODO: Simplify pipeline class by moving validation method here

from onemod.pipeline import Pipeline
from onemod.stage import Stage


def check_method(model: Pipeline | Stage, method: str) -> None:
    if isinstance(model, Pipeline):
        if method == "collect":
            raise ValueError(
                "Method 'collect' cannot be called on a pipeline instance"
            )
    else:
        if method == "collect":
            if model.has_submodels:
                if len(model.collect_after) == 0:
                    raise ValueError(
                        "Method 'collect' cannot be called on stage with empty collect_after"
                    )
            else:
                raise ValueError(
                    "Method 'collect' cannot be called on a stage without submodels"
                )
        elif method in model.skip:
            raise ValueError(
                f"Stage '{model.name}' skips the '{method}' method"
            )


def check_input_exists(
    model: Pipeline | Stage, stages: list[str] | None = None
) -> None:
    if isinstance(model, Pipeline):
        # Check input already exists for any upstream stage not being evaluated
        stages_being_evaluated = stages or model.stages.keys()
        for stage_name in stages_being_evaluated:
            if stage_name not in model.stages:
                raise ValueError(f"Stage '{stage_name}' not found in pipeline")
            stage = model.stages[stage_name]
            stage.input.check_exists(
                upstream_stages=[
                    dependency
                    for dependency in stage.dependencies
                    if dependency not in stages_being_evaluated
                ]
            )
    else:
        model.input.check_exists()


def collect_results(
    stage: Stage,
    method: str,
    subsets: dict | None,
    paramsets: dict | None,
    collect: bool | None,
) -> bool:
    """Determine whether to collect stage submodel results."""
    if stage.has_submodels:
        if method == "collect":
            return False
        if method in stage.collect_after:
            if collect is None:
                return subsets is None and paramsets is None
            else:
                return collect
        return False
    return False



from typing import Type

from pydantic import BaseModel, field_serializer, field_validator

from onemod.constraints.base import Constraint


class ColumnSpec(BaseModel):
    type: (
        Type[int]
        | Type[float]
        | Type[str]
        | Type[bool]
        | Type[BaseModel]
        | None
    ) = None
    constraints: list[Constraint] | None = None

    @field_serializer("type")
    def serialize_type(self, t, info):
        return t.__name__ if t else None

    @field_validator("type", mode="before")
    def deserialize_type(cls, v):
        if isinstance(v, str):
            type_mapping = {
                "int": int,
                "float": float,
                "str": str,
                "bool": bool,
            }
            if v in type_mapping:
                return type_mapping[v]
            else:
                raise ValueError(f"Unknown type string: {v}")
        return v

    @field_serializer("constraints")
    def serialize_constraints(self, constraints):
        if constraints is None:
            return None
        return [constraint.model_dump() for constraint in constraints]



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Literal, Type

import dill
import pandas as pd
import polars as pl
import tomli
import tomli_w
import yaml


class DataIO(ABC):
    """Bridge class that unifies the I/O for different data file types."""

    fextns: tuple[str, ...] = ("",)
    dtypes: tuple[Type, ...] = (pd.DataFrame, pl.DataFrame, pl.LazyFrame)

    def load_eager(
        self,
        fpath: Path | str,
        backend: Literal["pandas", "polars"] = "pandas",
        **options,
    ) -> pd.DataFrame | pl.DataFrame:
        """Load data from given path."""
        fpath = Path(fpath)
        if fpath.suffix not in self.fextns:
            raise ValueError(f"File extension must be in {self.fextns}.")
        if backend == "polars":
            return self._load_eager_polars_impl(fpath, **options)
        elif backend == "pandas":
            return self._load_eager_pandas_impl(fpath, **options)
        else:
            raise ValueError("Backend must be either 'polars' or 'pandas'.")

    def load_lazy(self, fpath: Path | str, **options) -> pl.LazyFrame:
        """Load data lazily (Polars only) from given path."""
        fpath = Path(fpath)
        if fpath.suffix not in self.fextns:
            raise ValueError(f"File extension must be in {self.fextns}.")
        return self._load_lazy_impl(fpath, **options)

    def dump(
        self,
        obj: pl.DataFrame | pl.LazyFrame | pd.DataFrame,
        fpath: Path | str,
        mkdir: bool = True,
        **options,
    ):
        """Dump data to given path."""
        fpath = Path(fpath)
        if mkdir:
            fpath.parent.mkdir(parents=True, exist_ok=True)

        if isinstance(obj, pl.LazyFrame):
            obj = obj.collect()

        if isinstance(obj, pl.DataFrame):
            self._dump_polars_impl(obj, fpath, **options)
        elif isinstance(obj, pd.DataFrame):
            self._dump_pandas_impl(obj, fpath, **options)
        else:
            raise TypeError(f"Data must be an instance of {self.dtypes}.")

    @abstractmethod
    def _load_eager_polars_impl(self, fpath: Path, **options) -> pl.DataFrame:
        """Polars implementation of eager loading."""
        pass

    @abstractmethod
    def _load_eager_pandas_impl(self, fpath: Path, **options) -> pd.DataFrame:
        """Pandas implementation of eager loading."""
        pass

    @abstractmethod
    def _load_lazy_impl(self, fpath: Path, **options) -> pl.LazyFrame:
        """Polars implementation of lazy loading."""
        pass

    @abstractmethod
    def _dump_polars_impl(self, obj: pl.DataFrame, fpath: Path, **options):
        """Polars implementation of dumping."""
        pass

    @abstractmethod
    def _dump_pandas_impl(self, obj: pd.DataFrame, fpath: Path, **options):
        """Pandas implementation of dumping."""
        pass

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"


class ConfigIO(ABC):
    """Bridge class that unifies the I/O for different config file types."""

    fextns: tuple[str, ...] = ("",)
    dtypes: tuple[Type, ...] = (object,)

    def load(self, fpath: Path | str, **options) -> Any:
        """Load data from given path."""
        fpath = Path(fpath)
        if fpath.suffix not in self.fextns:
            raise ValueError(f"File extension must be in {self.fextns}.")
        return self._load_impl(fpath, **options)

    def dump(self, obj: Any, fpath: Path | str, mkdir: bool = True, **options):
        """Dump data to given path."""
        fpath = Path(fpath)
        if not isinstance(obj, self.dtypes):
            raise TypeError(f"Data must be an instance of {self.dtypes}.")
        if mkdir:
            fpath.parent.mkdir(parents=True, exist_ok=True)
        self._dump_impl(obj, fpath, **options)

    @abstractmethod
    def _load_impl(self, fpath: Path, **options) -> Any:
        """Implementation of loading to be provided by subclasses."""
        pass

    @abstractmethod
    def _dump_impl(self, obj: Any, fpath: Path, **options):
        """Implementation of dumping to be provided by subclasses."""
        pass

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"


class CSVIO(DataIO):
    fextns: tuple[str, ...] = (".csv",)

    def _load_eager_polars_impl(self, fpath: Path, **options) -> pl.DataFrame:
        return pl.read_csv(fpath, **options)

    def _load_eager_pandas_impl(self, fpath: Path, **options) -> pd.DataFrame:
        return pd.read_csv(fpath, **options)

    def _load_lazy_impl(self, fpath: Path, **options) -> pl.LazyFrame:
        return pl.scan_csv(fpath, **options)

    def _dump_polars_impl(self, obj: pl.DataFrame, fpath: Path, **options):
        obj.write_csv(fpath, **options)

    def _dump_pandas_impl(self, obj: pd.DataFrame, fpath: Path, **options):
        obj.to_csv(fpath, index=False, **options)


class ParquetIO(DataIO):
    fextns: tuple[str, ...] = (".parquet",)

    def _load_eager_polars_impl(self, fpath: Path, **options) -> pl.DataFrame:
        return pl.read_parquet(fpath, **options)

    def _load_eager_pandas_impl(self, fpath: Path, **options) -> pd.DataFrame:
        return pd.read_parquet(fpath, **options)

    def _load_lazy_impl(self, fpath: Path, **options) -> pl.LazyFrame:
        return pl.scan_parquet(fpath, **options)

    def _dump_polars_impl(self, obj: pl.DataFrame, fpath: Path, **options):
        obj.write_parquet(fpath, **options)

    def _dump_pandas_impl(self, obj: pd.DataFrame, fpath: Path, **options):
        obj.to_parquet(fpath, index=False, **options)


class JSONIO(ConfigIO):
    fextns: tuple[str, ...] = (".json",)
    dtypes: tuple[Type, ...] = (dict, list)

    def _load_impl(self, fpath: Path, **options) -> dict | list:
        with open(fpath, "r") as f:
            return json.load(f, **options)

    def _dump_impl(self, obj: dict | list, fpath: Path, **options):
        with open(fpath, "w") as f:
            json.dump(obj, f, **options)


class PickleIO(ConfigIO):
    fextns: tuple[str, ...] = (".pkl", ".pickle")

    def _load_impl(self, fpath: Path, **options) -> Any:
        with open(fpath, "rb") as f:
            return dill.load(f, **options)

    def _dump_impl(self, obj: Any, fpath: Path, **options):
        with open(fpath, "wb") as f:
            return dill.dump(obj, f, **options)


class TOMLIO(ConfigIO):
    fextns: tuple[str, ...] = (".toml",)
    dtypes: tuple[Type, ...] = (dict,)

    def _load_impl(self, fpath: Path, **options) -> dict:
        with open(fpath, "rb") as f:
            return tomli.load(f, **options)

    def _dump_impl(self, obj: dict, fpath: Path, **options):
        with open(fpath, "wb") as f:
            tomli_w.dump(obj, f)


class YAMLIO(ConfigIO):
    fextns: tuple[str, ...] = (".yml", ".yaml")
    dtypes: tuple[Type, ...] = (dict, list)

    def _load_impl(self, fpath: Path, **options) -> dict | list:
        options = dict(Loader=yaml.SafeLoader) | options
        with open(fpath, "r") as f:
            return yaml.load(f, **options)

    def _dump_impl(self, obj: dict | list, fpath: Path, **options):
        options = dict(Dumper=yaml.SafeDumper) | options
        with open(fpath, "w") as f:
            return yaml.dump(obj, f, **options)


dataio_dict: dict[str, DataIO] = {
    fextn: io for io in [CSVIO(), ParquetIO()] for fextn in io.fextns
}

configio_dict: dict[str, ConfigIO] = {
    fextn: io
    for io in [JSONIO(), PickleIO(), TOMLIO(), YAMLIO()]
    for fextn in io.fextns
}



from pathlib import Path
from typing import Any, ClassVar

from polars import Boolean, DataFrame, Float64, Int64, String
from pydantic import BaseModel

from onemod.constraints import Constraint
from onemod.dtypes.column_spec import ColumnSpec
from onemod.fsutils.data_loader import DataLoader
from onemod.fsutils.io import configio_dict, dataio_dict
from onemod.validation.error_handling import (
    ValidationErrorCollector,
    handle_error,
)


class Data(BaseModel):
    stage: str | None = None
    methods: list[str] | None = None
    format: str | None = None
    path: Path | None = None
    shape: tuple[int, int] | None = None
    columns: dict[str, ColumnSpec] | None = None
    type_mapping: ClassVar[dict[type, Any]] = {
        bool: Boolean,
        int: Int64,
        float: Float64,
        str: String,
    }

    def model_post_init(self, *args, **kwargs) -> None:
        if self.format is None and self.path is None:
            raise ValueError("Data format and path cannot both be None")
        if self.path is not None:
            expected_format = self.path.suffix[1:] or "directory"
            if self.format is None:
                self.format = expected_format
            else:
                if self.format != expected_format:
                    raise ValueError("Data format and path do not match")

    def validate_metadata(
        self, kind: str, collector: ValidationErrorCollector | None = None
    ) -> None:
        """One-time validation for instance metadata."""
        if self.path is None:
            handle_error(
                self.stage,
                "Data validation",
                ValueError,
                "File path is required.",
                collector,
            )

        if self.format is not None:
            fextn = "." + self.format
            if fextn not in dataio_dict and fextn not in configio_dict:
                handle_error(
                    self.stage,
                    "Data validation",
                    ValueError,
                    f"Unsupported file format {self.format}.",
                    collector,
                )

        if self.shape:
            if not isinstance(self.shape, tuple) or len(self.shape) != 2:
                handle_error(
                    self.stage,
                    "Data validation",
                    ValueError,
                    "Shape must be a tuple of (rows, columns).",
                    collector,
                )

        if self.columns:
            for col_name, col_spec in self.columns.items():
                if (
                    hasattr(col_spec, "type")
                    and col_spec.type not in self.type_mapping
                ):
                    handle_error(
                        self.stage,
                        "Data validation",
                        ValueError,
                        f"Unsupported type {col_spec.type} for column {col_name}.",
                        collector,
                    )
                if hasattr(col_spec, "constraints") and col_spec.constraints:
                    for constraint in col_spec.constraints:
                        if not isinstance(constraint, Constraint):
                            handle_error(
                                self.stage,
                                "Data validation",
                                ValueError,
                                f"Invalid constraint specified for column {col_name}.",
                                collector,
                            )

    def validate_shape(
        self, data: DataFrame, collector: ValidationErrorCollector | None = None
    ) -> None:
        """Validate the shape of the data."""
        if data.shape != self.shape:
            handle_error(
                self.stage,
                "Data validation",
                ValueError,
                f"Expected DataFrame shape {self.shape}, got {data.shape}.",
                collector,
            )

    def validate_data(
        self,
        data: DataFrame | None,
        collector: ValidationErrorCollector | None = None,
    ) -> None:
        """Validate the columns and shape of the data."""
        if data is None and self.path is not None:
            try:
                data = DataLoader().load(self.path)
            except Exception as e:
                handle_error(
                    self.stage,
                    "Data validation",
                    e.__class__,
                    str(e),
                    collector,
                )
                return

        if self.shape:
            self.validate_shape(data, collector)

        if self.columns:
            self.validate_columns(data, collector)

    def validate_columns(
        self, data: DataFrame, collector: ValidationErrorCollector | None = None
    ) -> None:
        """Validate columns based on specified types and constraints."""
        if self.columns is None:
            return

        for col_name, col_spec in self.columns.items():
            if col_name not in data.columns:
                handle_error(
                    self.stage,
                    "Data validation",
                    ValueError,
                    f"Column '{col_name}' is missing from the data.",
                    collector,
                )

            expected_type = col_spec.type or None
            constraints = col_spec.constraints or []

            if expected_type:
                polars_type = self.type_mapping.get(
                    expected_type
                )  # TODO: better ways to handle this?
                if not polars_type:
                    handle_error(
                        self.stage,
                        "Data validation",
                        ValueError,
                        f"Unsupported type {expected_type}.",
                        collector,
                    )
                if data[col_name].dtype != polars_type:
                    handle_error(
                        self.stage,
                        "Data validation",
                        ValueError,
                        f"Column '{col_name}' must be of type {expected_type.__name__}.",
                        collector,
                    )

            for constraint in constraints:
                constraint.use_validation(data[col_name])
